local DGV = DugisGuideViewer
if not DGV then return end

local SmallFrame = DGV:RegisterModule("SmallFrame")--, "Guides")

SmallFrame.Frame = CreateFrame("Frame", "DugisSmallFrame", UIParent, "DugisSmallFrameTemplate")
SmallFrame.Frame:Hide()

function SmallFrame:Initialize()
	if SmallFrame.initialized then return end
	SmallFrame.initialized = true
	
	local L = DugisLocals

	local oldFrameSize = 0
	local oldTextSize = 0 
	local newFrameSize = 0
	local newTextSize = 0
	local flashGroup, flash
	
	local function IsFloating()
		return (not DugisGuideViewer:UserSetting(DGV_ANCHOREDSMALLFRAME)) or DugisGuideViewer.carboniteloaded or DugisGuideViewer.sexymaploaded or DugisGuideViewer.nuiloaded or not SmallFrame.loaded
	end
	
	local function GetNonTextWidth()
		local frame = DugisSmallFrame		
		local _, _, _, xOfs, yOfs = frame.Chk:GetPoint()
		local modelButtonWidth
		local currquest = DGV.CurrentQuestIndex
		
		if DGV:HasModel(currquest) then modelButtonWidth = frame.ModelButton:GetWidth( ) else modelButtonWidth = 4 end
		return frame.Waypoint:GetWidth() + frame.Icon:GetWidth() + frame.Chk:GetWidth() + modelButtonWidth + xOfs
	end

	local function WatchFrame_DisplaySmallFrame (lineFrame, nextAnchor, maxHeight, frameWidth)
		SmallFrame.Frame:RegisterForDrag(nil)
		SmallFrame.Frame:ClearAllPoints()	
		SmallFrame.Frame:SetBackdrop(nil)
		SmallFrame.Frame:SetPoint("RIGHT", WatchFrame, "RIGHT", 0, 0);
		SmallFrame.Frame:SetPoint("LEFT", WatchFrame, "LEFT", -40, 0);
		if (nextAnchor) then
			SmallFrame.Frame:SetPoint("TOP", nextAnchor, "BOTTOM", 0, -WATCHFRAME_TYPE_OFFSET);
		else
			SmallFrame.Frame:SetPoint("TOP", WatchFrameLines, "TOP", 0, -WATCHFRAME_INITIAL_OFFSET);
		end
		SmallFrame.Frame.Text:SetWordWrap(true)
		SmallFrame.Frame.Text:SetNonSpaceWrap(true)
		oldFrameSize, oldTextSize = SmallFrame.Frame:GetWidth(), SmallFrame.Frame:GetWidth() - GetNonTextWidth()
		newFrameSize, newTextSize = oldFrameSize, oldTextSize
		SmallFrame.Frame.Text:SetWidth(newTextSize)
		--SmallFrame.Frame.Text:SetHeight(29)
		SmallFrame.Frame:SetHeight(29)
		if SmallFrame.FlashFrame then
			SmallFrame.FlashFrame:ClearAllPoints()	
			SmallFrame.FlashFrame:SetPoint("CENTER", DGV.Modules.DugisWatchFrame.WatchBackground, 0, -1)
		end
		return SmallFrame.Frame, SmallFrame.Frame:GetWidth(), 1; 
	end

	function SmallFrame:Reset()
		if IsFloating()	then
			SmallFrame.Frame:ClearAllPoints()	
			SmallFrame.Frame:SetPoint("CENTER", 0, 220)
		end
		SmallFrame:ResetFloating()
	end

	local function UnsnapFromWatchFrame()
		local pointIndex
		for pointIndex=1, SmallFrame.Frame:GetNumPoints() do
			local _, relativeTo = SmallFrame.Frame:GetPoint(pointIndex)
			if relativeTo==WatchFrame then
				SmallFrame.Frame:RegisterForDrag("LeftButton")
				SmallFrame.Frame:ClearAllPoints()	
				SmallFrame.Frame:SetPoint("CENTER", 0, 220)
				newFrameSize, newTextSize = SmallFrame:GetCurrentWidth( )
				break;
			end
		end
	end

	function SmallFrame:ResetFloating()
		if IsFloating() or
			not DugisGuideViewer:UserSetting(DGV_SHOWSMALLFRAME) or 
			not DugisGuideViewer.GuideOn
		then
			SmallFrame.Frame.Text:SetWordWrap(false)
			SmallFrame.Frame.Text:SetNonSpaceWrap(false)
			SmallFrame.Frame:SetHeight(50)
			UnsnapFromWatchFrame()
			--SmallFrame.Frame.Text:SetHeight(50)
			--local fw, tw = SmallFrame:GetCurrentWidth()
			--DebugPrint("Debug ResetFloating: fw="..tostring(fw).." tw"..tostring(tw))
			SmallFrame:SetCurrentWidth(SmallFrame:GetCurrentWidth())
			DugisGuideViewer:SetSmallFrameBorder()
			if SmallFrame.FlashFrame then
				SmallFrame.FlashFrame:ClearAllPoints()	
				SmallFrame.FlashFrame:SetPoint("CENTER", SmallFrame.Frame, 1, -1)
			end
			DGV.Modules.DugisWatchFrame:RemoveObjectiveHandler(WatchFrame_DisplaySmallFrame)
		elseif DugisGuideViewer:UserSetting(DGV_SHOWSMALLFRAME) then
			DGV.Modules.DugisWatchFrame:AddObjectiveHandler(WatchFrame_DisplaySmallFrame)
		end
		WatchFrame_Update()
	end

	--[[
	local function PLAYER_REGEN_ENABLED(self)
		
			
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	end
	--]]

	function SmallFrame:OnClick(self, button)
		name = self:GetName()

		if name == "DugisSmallFrame" and button == "RightButton" then
			if DugisMainframe:IsVisible() == 1 then
				DugisGuideViewer:HideLargeWindow()
			else
				UIFrameFadeIn(DugisMainframe, 0.5, 0, 1)
				UIFrameFadeIn(Dugis, 0.5, 0, 1)
				DugisGuideViewer:ShowLargeWindow()
			end
		elseif button == "LeftButton" and IsShiftKeyDown() then
			local row = _G["DGVRow"..DGV.CurrentQuestIndex]
			DugisGuideViewer.Modules.StickyFrame:AddRow(row)
		end
	end

	--Return total width, text width
	function SmallFrame:GetCurrentWidth()

		local frame = DugisSmallFrame		
		local _, _, _, xOfs, yOfs = frame.Chk:GetPoint()
		local modelButtonWidth
		local currquest = DGV.CurrentQuestIndex
		
		if DGV:HasModel(currquest) then modelButtonWidth = frame.ModelButton:GetWidth( ) else modelButtonWidth = 4 end
		--width = Quest text + waypoint button + picture + chkbox + modelButton + left margin + right margin	
		--frame:SetWidth(frame.Text:GetWidth()  + frame.Waypoint:GetWidth() + frame.Icon:GetWidth() + frame.Chk:GetWidth() + modelButtonWidth + xOfs )
		--frame.ModelButton:SetPoint("LEFT", frame.Text, "RIGHT", 0, -1)
		local width = frame.Text:GetStringWidth()  + frame.Waypoint:GetWidth() + frame.Icon:GetWidth() + frame.Chk:GetWidth() + modelButtonWidth + xOfs
		return width, frame.Text:GetStringWidth()
	end

	function SmallFrame:SetCurrentWidth(frameWidth, textWidth)
		--DebugPrint("Debug SetCurrentWidth: frameWidth="..tostring(frameWidth).." textWidth="..tostring(textWidth).." stack="..debugstack())
		DugisSmallFrame:SetWidth(frameWidth)
		if textWidth then
			DugisSmallFrame.Text:SetWidth(textWidth)
		end
	end

	--Called possibly in combat
	function SmallFrame:PopulateSmallFrame()
		local currquest = DGV.CurrentQuestIndex
		if currquest and DGV.actions[currquest] and DGV:isValidGuide(CurrentTitle) == true then
			local frame = DugisSmallFrame
			local level = DGV:GetQuestLevel(DGV.qid[currquest])
			local qName = DGV.quests1L[currquest]
			local questpart = DGV:ReturnTag("QIDP", currquest)
			
			if (level and strmatch(DGV.actions[currquest], "[ACT]") and DGV:UserSetting(DGV_QUESTLEVELON)) or (level and questpart and strmatch(DGV.actions[currquest], "[NK]") and DGV:UserSetting(DGV_QUESTLEVELON)) then qName = "["..level.."] "..qName end

			frame.Text:SetText(qName)
			frame.Icon:SetNormalTexture(DGV:getIcon(DGV.actions[currquest], currquest))
			
			if DGV:HasCoord(currquest) then frame.Waypoint:Enable() else frame.Waypoint:Disable() end
			if DGV:HasModel(currquest) then frame.ModelButton:Show() else frame.ModelButton:Hide() end 	
			if DGV:ReturnTag("NT") then frame.Chk:Disable() else frame.Chk:Enable() end
			
			local color  = DGV:GetQuestDiffColor(currquest)

			if (strmatch(DGV.actions[currquest], "[ACT]") and color and DGV:UserSetting(DGV_QUESTCOLORON)) or (questpart and color and strmatch(DGV.actions[currquest], "[NK]") and DGV:UserSetting(DGV_QUESTCOLORON)) then	--set difficulty color on A/C/T actions
				frame.Text:SetTextColor(color.r, color.g, color.b, 1)
			else
				frame.Text:SetTextColor(1, 0.82, 0, 1) 
			end
			frame.ModelButton:SetPoint("LEFT", frame.Text, "RIGHT", 0, -1)
		end
		
	end
	
	function SmallFrame:CheckButton_OnEvent(self, event)
		local SmallFramebutton = _G["DugisSmallFrameChk"]

		if DGV.CurrentQuestIndex then --If a guide is loaded
			if DGV:ReturnTag("NT") then
				DugisSmallFrameChk:SetChecked(0)
			elseif DGV.CurrentQuestIndex == DGV:GetLastGuideNumRows() then--LastGuideNumRows then
				DugisSmallFrameChk:SetChecked(0)
				DugisGuideViewer:LoadNextGuide()
			else
				DugisGuideViewer:SetChkToComplete(DGV.CurrentQuestIndex)
				DelayandMoveToNextQuest(0.3)
			end
		end
	end

	local autoTooltipFadeTime = 2^1024
	local function ResetAutoTooltip()
		if SmallFrameTooltip then SmallFrameTooltip:SetAlpha(1) end
		autoTooltipFadeTime = math.huge
	end

	local function UpdateAutoTooltip()
		local toEnd = autoTooltipFadeTime-GetTime()
		--DebugPrint("Debug UpdateAutoTooltip: toEnd="..tostring(toEnd).." SmallFrameTooltip:GetAlpha()="..tostring(SmallFrameTooltip:GetAlpha()))
		if toEnd <= 0 then 
			SmallFrameTooltip:Hide()
			ResetAutoTooltip()
		elseif toEnd <= 3 then 
			SmallFrameTooltip:SetAlpha(toEnd/3) 
		end
	end

	local tooltip = CreateFrame( "GameTooltip", "SmallFrameTooltip", nil, "GameTooltipTemplate" ); 
	function SmallFrame:SmallWindowTooltip_OnEnter(self, event)
		if DugisGuideViewer:isValidGuide(CurrentTitle) == true and DGV.CurrentQuestIndex then

			local name = "DGVRow"..DGV.CurrentQuestIndex.."Desc"
			local text = _G[name]
			
			local statusFrameTooltipText = text:GetText()
			if statusFrameTooltipText and not DugisGuideViewer:GetDB(DGV_DISPLAYCOORDINATES) then
				statusFrameTooltipText = string.gsub(statusFrameTooltipText, "%([%d,.%s]*%)[,%s]*", "")
			end		
			
			--CreateFrame( "GameTooltip", "SmallFrameTooltip", nil, "GameTooltipTemplate" ); 
			SmallFrameTooltip:SetOwner( _G["DugisSmallFrame"])
			SmallFrameTooltip:SetFrameStrata("TOOLTIP") 
			SmallFrameTooltip:SetParent(UIParent)
			SmallFrameTooltipTextLeft1:SetFont("Fonts\\FRIZQT__.TTF", 12)
			SmallFrameTooltip:SetPadding(5)
			--SmallFrameTooltip:AddLine(text:GetText(), 1, 1, 1, 1,true)
			SmallFrameTooltip:AddLine(statusFrameTooltipText, 1, 1, 1, 1,true)
			SmallFrameTooltip:Show()

			local ttwidth, ttheight, fwidth, fheight, pad = DugisGuideViewer:GetToolTipSize(SmallFrameTooltip)
			--DebugPrint("fwidth:"..fwidth.." fheight:"..fheight.." ttwidth"..ttwidth.." ttheight"..ttheight.." pad"..pad)
			--[[
			local scaleFactor = fwidth / ttwidth
			local maxScale = 1.3

			if (scaleFactor > 1) then
				local newwidth
				if scaleFactor > maxScale then
					scaleFactor = maxScale
				end
				
				if(scaleFactor < 1.10) then
					newwidth = fwidth * 1.10
				else 
					newwidth = ttwidth * scaleFactor
				end
				SmallFrameTooltip:SetWidth(newwidth)
				SmallFrameTooltipTextLeft1:SetWidth(newwidth - 15)
				SmallFrameTooltip:SetHeight(SmallFrameTooltipTextLeft1:GetHeight() + 20)

			end
			--]]
				
			SmallFrameTooltip:ClearAllPoints()
			local anchorPoint = strupper(DugisGuideViewer:GetDB(DGV_TOOLTIPANCHOR)):gsub("%s", "")
			if anchorPoint=="DEFAULT" and IsFloating() then
				anchorPoint = "BOTTOMRIGHT"
			elseif anchorPoint=="DEFAULT" then
				anchorPoint = "TOPRIGHT"
			end
			
			local toolAnchorPoint 	= ""
			toolAnchorPoint = toolAnchorPoint..((anchorPoint:find("BOTTOM.*") and "TOP") or "")
			toolAnchorPoint	= toolAnchorPoint..((anchorPoint:find("TOP.*") and "BOTTOM") or "")
			toolAnchorPoint = toolAnchorPoint..((anchorPoint:find("^RIGHT") and "LEFT") or "")
			toolAnchorPoint = toolAnchorPoint..((anchorPoint:find("^LEFT") and "RIGHT") or "")
			toolAnchorPoint = toolAnchorPoint..((anchorPoint:find(".+LEFT") and "LEFT") or "")
			toolAnchorPoint = toolAnchorPoint..((anchorPoint:find(".+RIGHT") and "RIGHT") or "")
			
			
			--DebugPrint("Debug SmallWindowTooltip_OnEnter: toolAnchorPoint="..toolAnchorPoint.." anchorPoint="..anchorPoint)
			SmallFrameTooltip:SetPoint(toolAnchorPoint, DugisSmallFrame, anchorPoint, 0, 0)
			ResetAutoTooltip()
		end
	end
	
	function SmallFrame:OnDragStart(frame)
	  if not DGV:UserSetting(DGV_LOCKSMALLFRAME) then
		frame:StartMoving();
		frame.isMoving = true;
	  end
	end

	function SmallFrame:OnDragStop(self)
	  self:StopMovingOrSizing();
	  self.isMoving = false;
	end	
	
	function DugisGuideViewer:ShowAutoTooltip()
		--if 1 then return end
		if DugisGuideViewer:GetDB(DGV_SHOWTOOLTIP)==0 or (self.actions[CurrentQuestIndex] and strmatch(DugisGuideViewer.actions[CurrentQuestIndex], "[CNK]")==nil) or not DGV:UserSetting(DGV_SHOWSMALLFRAME) then
			if SmallFrameTooltip then SmallFrameTooltip:Hide() end
			ResetAutoTooltip()
			return 
		end
		SmallFrame:SmallWindowTooltip_OnEnter()
		autoTooltipFadeTime = GetTime() + DugisGuideViewer:GetDB(DGV_SHOWTOOLTIP) + 3
		--SmallFrameTooltip:Show()
		SmallFrameTooltip:SetScript("OnUpdate", UpdateAutoTooltip)
	end
	
	function SmallFrame:PlayFlashAnimation( )
		if not SmallFrame.FlashFrame then
			flashGroup, flash, SmallFrame.FlashFrame = DGV:CreateFlashFrame(SmallFrame.Frame)
			SmallFrame:ResetFloating()
		end
		
		if IsFloating() and DGV:UserSetting(DGV_SMALLFRAMEBORDER) then
			SmallFrame.FlashFrame:Show()
			SmallFrame.FlashFrame:SetWidth(SmallFrame.Frame:GetWidth() - 15)
			SmallFrame.FlashFrame:SetHeight(SmallFrame.Frame:GetHeight() - 18)
			flashGroup:Play()
		else
			SmallFrame.FlashFrame:Hide()
		end
	end

	local TransitionFont = nil
	local SmallFrameTransition = nil
	SmallFrame.Load = function()
		if SmallFrame.loaded then return end
		SmallFrame.loaded = true
		
		if DugisSmallFrameTransition then SmallFrameTransition = DugisSmallFrameTransition else
			SmallFrameTransition = CreateFrame("Frame", "DugisSmallFrameTransition", UIParent)
			SmallFrameTransition:SetHeight(32)
			SmallFrameTransition:SetWidth(100)
		end
		SmallFrameTransition:Show()

		
		if SmallFrameTransitionFont then TransitionFont = SmallFrameTransitionFont else
			TransitionFont = SmallFrameTransition:CreateFontString("SmallFrameTransitionFont", "ARTWORK", "GameFontNormal")
			TransitionFont:SetPoint("LEFT")
		end
		--TransitionFont:Show()

		local smallElapsed = 1.5
		local watchElapsed = -1
		local suspendFrameTransition = nil
		SmallFrameTransition:SetScript("OnUpdate", function(self, el)
			smallElapsed = smallElapsed + el
			--[[if watchElapsed>=0 then
				watchElapsed = watchElapsed + el
				if watchElapsed > 1 then
					if not IsFloating() then
						if newWatchFrameHeight then
							WatchFrame:SetHeight(newWatchFrameHeight)
						end
						WatchFrame_Update()
						UnsnapWatchFrame()
					end
					watchElapsed = -1
					suspendFrameTransition = nil
				else
					suspendFrameTransition = true
					if not IsFloating() and oldWatchFrameHeight and newWatchFrameHeight then
						local wfw = math.ceil(oldWatchFrameHeight + 
							(newWatchFrameHeight-oldWatchFrameHeight)*watchElapsed)
						WatchFrame:SetHeight(wfw)
					end
				end
			end]]
			--DebugPrint("smallElapsed="..smallElapsed)
			if smallElapsed > 1 then
				SmallFrame:SetCurrentWidth(newFrameSize, newTextSize)
			else
				self:SetPoint("LEFT", DugisSmallFrame.Icon, "RIGHT", 0, smallElapsed*40)
				self:SetAlpha(1-smallElapsed)
				local w1 = oldFrameSize + (newFrameSize-oldFrameSize)*smallElapsed
				local w2 = oldTextSize + (newTextSize-oldTextSize)*smallElapsed
				SmallFrame:SetCurrentWidth(w1, w2)
			end
			
			if watchElapsed < 0 and smallElapsed > 1 then
				self:Hide()
				self:SetAlpha(0)
			end
		end)

		--[[function SmallFrame:StartWatchFrameTransition( )
			if suspendFrameTransition then return end

			watchElapsed = 0
			--SmallFrameTransition:SetAlpha(1)
			SmallFrameTransition:Show()
		end]]

		function SmallFrame:StartFrameTransition( )
			--SmallFrame:StartWatchFrameTransition()
			local fontObj = DugisSmallFrameText:GetFontObject()
			local textR, textG, textB = DugisSmallFrameText:GetTextColor()
			
			SmallFrameTransitionFont:SetFontObject(fontObj)
			SmallFrameTransitionFont:SetTextColor(textR, textG, textB)
			SmallFrameTransitionFont:SetText(DugisSmallFrame.Text:GetText())
			SmallFrameTransition:SetPoint("LEFT", DugisSmallFrame.Icon, "RIGHT", 0, 0)
			
			smallElapsed = 0
			SmallFrameTransition:SetAlpha(1)
			SmallFrameTransition:Show()
		end
		
		function SmallFrame:Init( )
			--SmallFrame.Frame = DugisSmallFrame
			SmallFrame.Frame:SetScript("OnEvent", PLAYER_REGEN_ENABLED)
			SmallFrame.Frame.Text:SetWordWrap(false)
		end
		
		function SmallFrame:Enable()
			if DGV:UserSetting(DGV_SHOWSMALLFRAME) then
				SmallFrame.Frame:Show()
				SmallFrameTransition:Show()
				SmallFrameTooltip:Show()
				SmallFrame:ResetFloating()
			end
		end

		function SmallFrame:Disable()
			SmallFrame.Frame:Hide()
			SmallFrameTransition:Hide()
			SmallFrameTooltip:Hide()
			SmallFrame:ResetFloating()
		end
		
		function DGV:LoadInitialView(text, texture)
			SmallFrame.Frame.Text:SetText(text)
			SmallFrame.Frame.Icon:SetNormalTexture(texture)
			SmallFrame:SetCurrentWidth(SmallFrame:GetCurrentWidth())
		end
		
		function DGV:UpdateSmallFrame()
			DGV:DebugFormat("UpdateSmallFrame")
			if DGV:UserSetting(DGV_SHOWSMALLFRAME) then
				SmallFrame.Frame:Show()
				if DGV:UserSetting(DGV_SMALLFRAMETRANSITION) == L["Scroll"] then
					oldFrameSize, oldTextSize = SmallFrame:GetCurrentWidth( )
					SmallFrame:StartFrameTransition( )
					SmallFrame:PopulateSmallFrame( )
					newFrameSize, newTextSize = SmallFrame:GetCurrentWidth( )
					--UIFrameFadeIn(SmallFrame.Frame.Text, 0.8, 0, 1)
					SmallFrame:ResetFloating()
				else
					SmallFrame:PopulateSmallFrame( )
					newFrameSize, newTextSize = SmallFrame:GetCurrentWidth( )
					--SmallFrame:SetCurrentWidth(newFrameSize, newTextSize)
					--UIFrameFadeIn(SmallFrame.Frame, 0.8, 0, 1)
					SmallFrame:ResetFloating()
					SmallFrame:PlayFlashAnimation( )	
				end
				DugisGuideViewer:ShowAutoTooltip()
			end
		end

		function DGV:OnWatchFrameUpdate()
			if DGV.Modules.DugisWatchFrame:ShouldModWatchFrame() and
				DugisGuideViewer:UserSetting(DGV_SHOWSMALLFRAME) and
				not WatchFrame.collapsed
			then
				SmallFrame.Frame:Show()
			else
				SmallFrameTooltip:Hide()
				if not IsFloating() then
					SmallFrame.Frame:Hide()
				end
			end
		end
		
		SmallFrame:Init()
		SmallFrame:Enable()
	end
	SmallFrame.Unload = function()
		if not SmallFrame.loaded then return end
		SmallFrame.loaded = false
		SmallFrame:Reset()
		SmallFrame:Disable()
		SmallFrame.Frame:Hide()
		SmallFrameTransition = nil
	end
end