local DGV = DugisGuideViewer
if not DGV then return end
local L = DugisLocals

local WMT = DGV:RegisterModule("WorldMapTracking")
WMT.essential = true

function WMT:Initialize()

	--[[Dump: value=GetTrackingInfo(1)
	[1]="Auctioneer",
	[2]="Interface\\Minimap\\Tracking\\Auctioneer",
	[4]="other",
	[5]=2
	Dump: value=GetTrackingInfo(2)
	[1]="Banker",
	 [2]="Interface\\Minimap\\Tracking\\Banker",
	 [4]="other",
	 [5]=2
	Dump: value=GetTrackingInfo(3)
	[1]="Battlemaster",
	 [2]="Interface\\Minimap\\Tracking\\BattleMaster",
	 [4]="other",
	 [5]=2
	Dump: value=GetTrackingInfo(4)
	[1]="Class Trainer",
	 [2]="Interface\\Minimap\\Tracking\\Class",
	 [3]=1,
	 [4]="other",
	 [5]=2
	Dump: value=GetTrackingInfo(5)
	[1]="Flight Master",
	 [2]="Interface\\Minimap\\Tracking\\FlightMaster",
	 [4]="other",
	 [5]=-1
	Dump: value=GetTrackingInfo(6)
	[1]="Food & Drink",
	 [2]="Interface\\Minimap\\Tracking\\Food",
	 [4]="other",
	 [5]=2
	Dump: value=GetTrackingInfo(7)
	[1]="Innkeeper",
	 [2]="Interface\\Minimap\\Tracking\\Innkeeper",
	 [4]="other",
	 [5]=2
	Dump: value=GetTrackingInfo(8)
	[1]="Mailbox",
	 [2]="Interface\\Minimap\\Tracking\\Mailbox",
	 [3]=1,
	 [4]="other",
	 [5]=2
	Dump: value=GetTrackingInfo(8)
	[1]="Mailbox",
	 [2]="Interface\\Minimap\\Tracking\\Mailbox",
	 [3]=1,
	 [4]="other",
	 [5]=2
	Dump: value=GetTrackingInfo(9)
	[1]="Poisons",
	 [2]="Interface\\Minimap\\Tracking\\Poisons",
	 [3]=1,
	 [4]="other",
	 Dump: value=GetTrackingInfo(10)
	[1]="Profession Trainers",
	 [2]="Interface\\Minimap\\Tracking\\Profession",
	 [3]=1,
	 [4]="other",
	 [5]=2
	Dump: value=GetTrackingInfo(11)
	[1]="Reagents",
	 [2]="Interface\\Minimap\\Tracking\\Reagents",
	 [4]="other",
	 [5]=2
	Dump: value=GetTrackingInfo(12)
	[1]="Repair",
	 [2]="Interface\\Minimap\\Tracking\\Repair",
	 [4]="other",
	 [5]=2]]

	local trackingIndex =
	{
		["Interface\\Minimap\\Tracking\\Auctioneer"] = 1,
		["Interface\\Minimap\\Tracking\\Banker"] = 2,
		["Interface\\Minimap\\Tracking\\BattleMaster"] = 3,
		["Interface\\Minimap\\Tracking\\Class"] = 4,
		["Interface\\Minimap\\Tracking\\FlightMaster"] = 5,
		["Interface\\Minimap\\Tracking\\Food"] = 6,
		["Interface\\Minimap\\Tracking\\Innkeeper"] = 7,
		["Interface\\Minimap\\Tracking\\Mailbox"] = 8,
		["Interface\\Minimap\\Tracking\\Poisons"] = 9,
		["Interface\\Minimap\\Tracking\\Profession"] = 10,
		["Interface\\Minimap\\Tracking\\Reagents"] = 11,
		["Interface\\Minimap\\Tracking\\Repair"] = 12
	}

	local trackingMap = {}
	local real_GetTrackingInfo = GetTrackingInfo
	do
		local i;
		for i=1,GetNumTrackingTypes() do
			local _,icon,_,info = real_GetTrackingInfo(i)
			
			if trackingIndex[icon] then
				trackingMap[trackingIndex[icon]] = i
			end
		end
	end
	
	local function GetTrackingInfo(id)
		return real_GetTrackingInfo(trackingMap[id])
	end

	local real_GetMapInfo = GetMapInfo
	local function GetMapInfo()
		local mapFileName, textureHeight, textureWidth = real_GetMapInfo()
		if not mapFileName then return mapFileName, textureHeight, textureWidth end
		return string.match(real_GetMapInfo(), "[^_]*"), textureHeight, textureWidth
	end
	

	local function IterateTrackingTypes()
		local count = GetNumTrackingTypes();
		local id = 0
		return function()
			id = id + 1
			if id<=count then return id, GetTrackingInfo(id) end --name, texture, active, category
		end
	end

	local professionTable = setmetatable({},
	{
		__index = function(t,i)
			local spell = tonumber(i)
			local v = i
			if spell then
				v = (GetSpellInfo(i))
			end
			return L[v]
		end,
	})

	local englishProfessionTable= setmetatable(
	{
		["2259"]	= "Alchemy",
		["3100"]	= "Blacksmithing",
		["7411"]	= "Enchanting",
		["4036"]	= "Engineering",
		["45357"]	= "Inscription",
		["25229"]	= "Jewelcrafting",
		["2108"]	= "Leatherworking",
		["3908"]	= "Tailoring",
		["9134"]	= "Herbalism",
		["2575"]	= "Mining",
		["8613"]	= "Skinning",
		["2550"]	= "Cooking",
		["3273"]	= "First Aid",
		["7620"]	= "Fishing",
	},
	{__index=function(t,k) rawset(t, k, k); return k; end})

	local DataProviders = {
		--["Mailbox"] = {},
		["Vendor"] = {},
		["ClassTrainer"] = {},
		["ProfessionTrainer"] = {},
		["Banker"] = {},
		["Battlemaster"] = {}
	}
	WMT.DataProviders = DataProviders

	function DataProviders:SelectProvider(trackingType, location, ...)
		for k,v in pairs(DataProviders) do
			if type(v)=="table" and v.ProvidesFor and v:ProvidesFor(trackingType, location, ...) then
				return v
			end
		end
	end
	
	function DataProviders:GetTooltipText(provider, trackingType, location, ...)
		provider = provider or self:SelectProvider(trackingType, location, ...)
		if provider and provider.GetTooltipText then
			return provider:GetTooltipText(trackingType, location, ...)
		else
			return (GetTrackingInfo(trackingType))
		end
	end

	function DataProviders:ShouldShow(provider, trackingType, location, ...)
		provider = provider or self:SelectProvider(trackingType, location, ...)
		if provider and provider.ShouldShow then
			return provider:ShouldShow(trackingType, location, ...)
		else
			return true
		end
	end

	local function GetNPCTT1(trackingType, location, npc)
		if DGV.GetLocalizedNPC then
			return DGV:GetLocalizedNPC(npc)
		end
	end

	function DataProviders.Vendor:ProvidesFor(trackingType)
		return trackingType==1 or trackingType==5 or trackingType==6  or trackingType==7 or
			trackingType==9 or trackingType==11 or trackingType==12
	end

	function DataProviders.Vendor:ShouldShow(trackingType)
		local class = select(2,UnitClass("player"))
		if (trackingType==9 and class~="ROGUE") then return false end
		return true
	end

	function DGV:GetFlightMasterName(npc)
		return DataProviders.Vendor:GetTooltipText(5, nil, npc)
	end

	function DataProviders.Vendor:GetTooltipText(trackingType, ...)
		return GetNPCTT1(trackingType, ...) or (GetTrackingInfo(trackingType)) 
	end

	function DataProviders.ClassTrainer:ProvidesFor(trackingType)
		return trackingType==4
	end

	local function GetGildedNPCTooltip(guildFunc, ...)
		local tt1 = GetNPCTT1(...)
		local tt2;
		if tt1 then tt2 = "<"..guildFunc(...)..">" end
		return tt1 or guildFunc(...), tt2
	end
	
	function DataProviders.ClassTrainer:GetTooltipText(trackingType, location, npc, class, gender)
		local genderString = ""
		if gender=="F" then genderString=" Female" end
		return GetGildedNPCTooltip(
			function(trackingType, location, npc, class) return L[class.." Trainer"..genderString] end,
					trackingType, location, npc, class, gender)
	end

	function DataProviders.ClassTrainer:ShouldShow(trackingType, location, npc, class)
		return class:lower()==select(2,UnitClass("player")):lower()
	end

	function DataProviders.ProfessionTrainer:ProvidesFor(trackingType)
		return trackingType==10
	end
	
	function DataProviders.ProfessionTrainer:GetTooltipText(trackingType, location, npc, spell, gender)
		local genderString = ""
		if gender=="F" then genderString=" Female" end
		return GetGildedNPCTooltip(
			function(trackingType, location, npc, spell) return L[englishProfessionTable[spell].." Trainer"..genderString] end,
					trackingType, location, npc, spell)
	end
	
	function DataProviders.ProfessionTrainer:ShouldShow(trackingType, location, npc, spell)
		local spellNum = tonumber(spell)
		local class = select(2,UnitClass("player"))
		if (spell=="Portal" and class~="MAGE") or
			(spell=="Pet" and class~="HUNTER")
		then return false end
		--[[if not spellNum then return true end
		local prof1, prof2 = GetProfessions()
		return (not prof1) or (not prof2) or --unchosen professions
			spellNum==2550 or spellnum==3273 or spellNum==7620 or --cooking,first aid,fishing,
			IsUsableSpell(GetSpellInfo(spellNum))]]
		return true
	end

	function DataProviders.Banker:ProvidesFor(trackingType)
		return trackingType==2
	end
	
	function DataProviders.Banker:GetTooltipText(...)
		return GetGildedNPCTooltip(
			function(...) return L["Banker"] end, ...)
	end

	function DataProviders.Battlemaster:ProvidesFor(trackingType)
		return trackingType==3
	end
	
	function DataProviders.Battlemaster:GetTooltipText(...)
		return GetGildedNPCTooltip(
			function(...) return L["Battlemaster"] end, ...)
		end

	function DGV:UnpackXY(coord)
		if not tonumber(coord) then
		  DGV:DebugFormat("UnpackXY", "coord", coord, "stack", debugstack())
		end
		local factor = 2^16
		local x,y =  floor(coord / factor) / factor, (coord % factor) / factor
		--DGV:DebugFormat("GetXY", "x", x, "y", y)
		return x,y
	end

	--local trackingStates = nil
	local trackingPoints = {}
	WMT.trackingPoints = trackingPoints
	local function UpdateTrackingFilters()
		local mapID, level = GetCurrentMapAreaID(), GetCurrentMapDungeonLevel()
		for _,point in ipairs(trackingPoints) do
			local trackingType, coord = unpack(point.args)
			local _, icon, active = GetTrackingInfo(trackingType)
			if active and DataProviders:ShouldShow(point.provider, unpack(point.args)) then
				if not point:IsShown() then
					point.icon:SetTexture(icon)
					point:SetHeight(12)
					point:SetWidth(12)
					point:Show()
					DGV:PlaceIconOnWorldMap(WorldMapButton, point, mapID, level, DGV:UnpackXY(coord) )
				end
			else
				point:Hide()
			end
		end
	end

	hooksecurefunc("SetTracking", UpdateTrackingFilters)
	hooksecurefunc("ClearAllTracking", UpdateTrackingFilters)

	local function AddWaypoint(point)
		local x, y = DGV:UnpackXY(point.args[2])
		DGV:AddCustomWaypoint(
			x, y, DataProviders:GetTooltipText(point.provider, unpack(point.args)),
			GetCurrentMapAreaID(), GetCurrentMapDungeonLevel())
	end

	local function point_OnClick(self, button)
		if button == "RightButton" then
			local menu = DGV.ArrowMenu:CreateMenu()
			DGV.ArrowMenu:CreateMenuTitle(menu,
					DataProviders:GetTooltipText(self.provider, unpack(self.args)))
			local setWay = DGV.ArrowMenu:CreateMenuItem(menu, L["Set as waypoint"])
			setWay:SetFunction(function ()
				DGV:RemoveAllWaypoints()
				AddWaypoint(self)
			end)
			local addWay = DGV.ArrowMenu:CreateMenuItem(menu, L["Add waypoint"])
			addWay:SetFunction(function () AddWaypoint(self)  end)
				local untrack = DGV.ArrowMenu:CreateMenuItem(menu,
					string.format(L["Remove %s Tracking"], (GetTrackingInfo(self.args[1]))))
			untrack:SetFunction(function ()
				MiniMapTracking_SetTracking(nil, self.args[1], nil, false)
			end)
			menu:ShowAtCursor()
		elseif button == "LeftButton" then
			if not IsShiftKeyDown() then
				DGV:RemoveAllWaypoints()
			end
			AddWaypoint(self)
		end
	end

	local function point_OnEnter(self, button)
		if UIParent:IsVisible() then
			DugisWaypointTooltip:SetParent(UIParent)
		else
			DugisWaypointTooltip:SetParent(self)
		end

		DugisWaypointTooltip:SetOwner(self, "ANCHOR_BOTTOMLEFT")
		local tt1, tt2 = DataProviders:GetTooltipText(self.provider, unpack(self.args))
		DugisWaypointTooltip:AddLine(tt1)
		if tt2 then DugisWaypointTooltip:AddLine(tt2) end
		DugisWaypointTooltip:Show()
	end

	local function point_OnLeave(self, button)
		DugisWaypointTooltip:Hide()
	end

	local trackingPointPool = {}
	local function GetCreatePoint(...)
		local point = tremove(trackingPointPool)
		if not point then
			point = CreateFrame("Button", nil, DugisMapOverlayFrame)
			point:RegisterForClicks("RightButtonUp","LeftButtonUp")
			point:SetScript("OnClick", point_OnClick)
			point:SetScript("OnEnter", point_OnEnter)
			point:SetScript("OnLeave", point_OnLeave)
			point.icon = point:CreateTexture("ARTWORK")
			point.icon:SetAllPoints()
			point.icon:Show()
		end
		point:Hide()
		point.args = {...}
		point.args[1] = tonumber(point.args[1])
		point.args[2] = tonumber(point.args[2])
		point.provider = DataProviders:SelectProvider(...)
		tinsert(trackingPoints, point)
		
		return point
	end

	local function GetDistance(point)
		local DugisArrow = DGV.Modules.DugisArrow
		--local x, y = GetXY(point.args[2])
		local x, y = DGV:UnpackXY(point[4])
		--return DGV:ComputeDistance(GetCurrentMapAreaID(), GetCurrentMapDungeonLevel(), x, y,
		--	DugisArrow.map, DugisArrow.floor, DugisArrow.pos_x, DugisArrow.pos_y)
		return DGV:ComputeDistance(point[1], point[2] or  DugisArrow.floor, x, y,
			DugisArrow.map, DugisArrow.floor, DugisArrow.pos_x, DugisArrow.pos_y)
	end

	local function IterateZonePoints(mapName, pointData)
		if not pointData then return end
		local DugisArrow = DGV.Modules.DugisArrow
		local currentSystem =  DGV.astrolabe:GetMapInfo(DugisArrow.map)
		local map,level = strsplit(":",mapName)
		map = DGV:GetMapIDFromName(map)
		level = tonumber(level)
		if not currentSystem==DGV.astrolabe:GetMapInfo(map, level) then
			return
		end
		local index = 0
		local zonePointIterator
		zonePointIterator = function()
			index = index + 1
			if not pointData[index] then return nil end
			local point = {map, level, strsplit(":", pointData[index])}
			point[3] = tonumber(point[3])
			point[4] = tonumber(point[4])
			if DataProviders:ShouldShow(nil, point[3], point[4], unpack(point, 5)) and
				GetDistance(point)
			then
				return point
			else
				return zonePointIterator()
			end
		end
		return zonePointIterator
	end

	function IterateAllFindNearestPoints()
		local faction = UnitFactionGroup("player")
		local opposingFaction
		if faction=="Alliance" then opposingFaction = "Horde"
		else opposingFaction = "Alliance"
		end

		local key, value, factionTable, factionKey, zonePointIterator
		local rootIterator
		rootIterator = function()
			if zonePointIterator then
				local tmp = zonePointIterator()
				if tmp then
					return tmp
				else
					zonePointIterator=nil
				end
			end
			if factionTable then
				factionKey, value = next(factionTable, factionKey)
				if factionKey then
					zonePointIterator = IterateZonePoints(factionKey, value)
				else
					factionTable = nil
				end
			else
				key,value = next(DugisWorldMapTrackingPoints, key)
				if not key then return nil
				elseif key==faction then
					factionTable = value
				elseif key~=opposingFaction then
					zonePointIterator = IterateZonePoints(key, value)
				end
			end
			return rootIterator()
		end
		return rootIterator
	end

	local function RemovePoint(point)
		local val, index
		for index, val in ipairs(trackingPoints) do
			if point == val then
				point:Hide()
				tinsert(trackingPointPool, tremove(trackingPoints, index))
				return
			end
		end
	end

	local function RemoveAllPoints()
		while #trackingPoints>0 do
			local point = tremove(trackingPoints)
			point:Hide()
			tinsert(trackingPointPool, point)
		end
	end

	local function AddPointData(pointData)
		if not pointData then return end
		local data
		for _,data in ipairs(pointData) do
			GetCreatePoint(strsplit(":", data))
		end
	end

	function WMT:UpdateTrackingMap()
		if not WMT.loaded then return end
		local mapName, level = GetMapInfo(), GetCurrentMapDungeonLevel()
		RemoveAllPoints()
		if not mapName or not DGV:UserSetting(DGV_WORLDMAPTRACKING) then return end
		AddPointData(DugisWorldMapTrackingPoints[UnitFactionGroup("player")][mapName..":"..level]);
		AddPointData(DugisWorldMapTrackingPoints[mapName])

		UpdateTrackingFilters()
	end

	--[[local function ClearAllTrackingFilters()
		local id, name, texture, active, category;
		for id, active in pairs(trackingStates) do
			trackingStates[id] = false
		end
		UpdateTrackingFilters()
	end

	function TrackingDropDown_IsNoTrackingActive()
		local id, name, texture, active, category;
		for id, active in pairs(trackingStates) do
			if (active) then
				return false;
			end
		end
		return true;
	end]]

	local function TrackingDropDown_Initialize(self, level)
		local id, name, texture, active, category, nested;
		local count = GetNumTrackingTypes();
		local info;
		
		if (level == 1) then
			info = UIDropDownMenu_CreateInfo();
			info.text=MINIMAP_TRACKING_NONE;
			info.checked = MiniMapTrackingDropDown_IsNoTrackingActive;
			info.func = ClearAllTracking;
			info.icon = nil;
			info.arg1 = nil;
			info.isNotRadio = true;
			info.keepShownOnClick = true;
			UIDropDownMenu_AddButton(info, level);
		end

		 
		for id, name, texture, active, category, nested in IterateTrackingTypes() do
			if id > 12 then break end
			info = UIDropDownMenu_CreateInfo();
			info.text = name;
			info.checked = MiniMapTrackingDropDownButton_IsActive
			info.func = MiniMapTracking_SetTracking;
			info.icon = texture;
			info.arg1 = id;
			info.isNotRadio = true;
			info.keepShownOnClick = true
			
			info.tCoordLeft = 0;
			info.tCoordRight = 1;
			info.tCoordTop = 0;
			info.tCoordBottom = 1;
			
			--if id~=5 then --skip flightmaster for now
				if level == 1 and nested < 0 then -- this tracking shouldn't be nested
					UIDropDownMenu_AddButton(info, level)
				elseif (level == 1 and (nested == TOWNSFOLK ) --[[and nested == UIDROPDOWNMENU_MENU_VALUE]]) then
					UIDropDownMenu_AddButton(info, level);
				end
			--end
		end
	end

	local function GetNearest(button)
		local shortest, shortestDist
		--for _,point in ipairs(trackingPoints) do
		--	local selected
		--	if (button.arg1 and button.arg1==point.args[4]) or button.arg1==point.args[1] then
		for point in IterateAllFindNearestPoints() do
			local selected
			if (button.arg1 and button.arg1==point[6]) or button.arg1==point[3] then
				selected = point
				local dist = GetDistance(selected)
				--DGV:DebugFormat("GetNearest","point", point, "dist", dist, "shortestDist", shortestDist)
				if dist and (not shortestDist or dist < shortestDist) then
					shortest = selected
					shortestDist = dist
				end
			end
		end
		return shortest
	end

	local function FindNearest(button)
		local DugisArrow = DGV.Modules.DugisArrow
		DGV:RemoveAllWaypoints()
		--AddWaypoint(GetNearest(button))
		local nearest = GetNearest(button)
		if nearest then
			local x, y = DGV:UnpackXY(nearest[4])
			local map, level = nearest[1], nearest[2] or DugisArrow.floor
			DGV:AddCustomWaypoint(
				x, y, DataProviders:GetTooltipText(nil, unpack(nearest, 3)),
				map, level)
		end
	end

	local function IterateDropdownLevel(level)
		local listFrame = _G["DropDownList"..level];
		local listFrameName = listFrame:GetName();
		local count = listFrame.numButtons
		local i = 0
		return function()
			i = i + 1
			if i<=count then return _G[listFrameName.."Button"..i] end
		end
	end

	local function Get_FindNearestDropDown_Initialize(baseLevel, value)
		return function (self, level)
			if level==baseLevel and UIDROPDOWNMENU_MENU_VALUE==value then
				--for _,point in ipairs(trackingPoints) do
					--if DataProviders:ShouldShow(point.provider, unpack(point.args)) then
						--DGV:DebugFormat("Get_FindNearestDropDown_Initialize", "point", point)
						--local trackingType = point.args[1]
				for point in IterateAllFindNearestPoints() do
					local button
					local found = false
					local trackingType = point[3]
					local name, texture = GetTrackingInfo(trackingType)
					for button in IterateDropdownLevel(level) do
						if name==button.value then
							found = true
							break;
						end
					end
					--DGV:DebugFormat("Get_FindNearestDropDown_Initialize", "found", found)
					if not found then
						local info;
						info = UIDropDownMenu_CreateInfo();
						info.text = name
						--info.checked = MiniMapTrackingDropDownButton_IsActive
						info.func = FindNearest;
						info.icon = texture;
						info.arg1 = trackingType;
						info.isNotRadio = true;
						info.notCheckable = true
						info.keepShownOnClick = true;
						info.tCoordLeft = 0;
						info.tCoordRight = 1;
						info.tCoordTop = 0;
						info.tCoordBottom = 1;

						if trackingType==10 then
							info.icon = nil;
							info.func =  nil;
							info.notCheckable = true;
							info.keepShownOnClick = false;
							info.hasArrow = true;
						end
						--if trackingType~=5 then --skip flightmaster for now
							UIDropDownMenu_AddButton(info, level)
						--end
					end
				end
			end

			if level==baseLevel+1 then
				--for _,point in ipairs(trackingPoints) do
				--	local trackingType, _, _, spell = unpack(point.args)
				for point in IterateAllFindNearestPoints() do
					local trackingType, _, _, spell = unpack(point, 3)
					if trackingType==10 then
						local button
						local found = false
						for button in IterateDropdownLevel(level) do
							if spell==button.arg1 then
								found = true
								break;
							end
						end
						if not found then
							local info;
							info = UIDropDownMenu_CreateInfo();
							info.text = professionTable[spell];
							--info.checked = MiniMapTrackingDropDown_IsNoTrackingActive;
							info.func = FindNearest;
							info.notCheckable = true
							info.icon = select(2, GetTrackingInfo(10));
							info.arg1 = spell;
							info.isNotRadio = true;
							info.keepShownOnClick = false;
							UIDropDownMenu_AddButton(info, level);
						end
					end
				end
			end
		end
	end

	--[[function WMT:ShowMenu(frame)
		local menu = DugisGuideViewer.ArrowMenu:CreateMenu()
		DugisGuideViewer.ArrowMenu:CreateMenuTitle(menu, L["World Map Tracking"])
		
		local filters = DugisGuideViewer.ArrowMenu:CreateMenuItem(menu, L["Set Tracking Filters..."])
		filters:SetFunction(function () ToggleDropDownMenu(1, nil, WorldMapTrackingDropDown, menu, 0, -5); end)
			
		--DGV:DebugFormat("ShowMenu", "menu:GetName", menu:GetName())
		if not WorldMapTrackingDropDown then
			CreateFrame("Frame", "WorldMapTrackingDropDown", UIParent, "UIDropDownMenuTemplate")
		end
		UIDropDownMenu_Initialize(WorldMapTrackingDropDown, TrackingDropDown_Initialize, "MENU")
		
		local nearest = DugisGuideViewer.ArrowMenu:CreateMenuItem(menu, L["Find Nearest..."])
		nearest:SetFunction(function () ToggleDropDownMenu(1, nil, FindNearestDropDown, menu, 0, -5); end)
	
		if not FindNearestDropDown then
			CreateFrame("Frame", "FindNearestDropDown", UIParent, "UIDropDownMenuTemplate")
		end
		UIDropDownMenu_Initialize(FindNearestDropDown, Get_FindNearestDropDown_Initialize(1), "MENU")

		menu:ShowAtCursor()
		
		return true
	end]]

	function DGV:WORLD_MAP_UPDATE()
		WMT:UpdateTrackingMap()
	end

	local orig_MiniMapTrackingDropDown_Initialize
	function WMT:Load()
		--[[if not trackingStates then
			trackingStates = {}
			local id, name, texture, active, category;
			for id, name, texture, active, category in IterateTrackingTypes() do
				trackingStates[id] = active
			end
		end]]
		orig_MiniMapTrackingDropDown_Initialize = MiniMapTrackingDropDown.initialize
		MiniMapTrackingDropDown.initialize = function(self, level)
			--DGV:DebugFormat("MiniMapTrackingDropDown_Initialize")
			if not WorldMapButton:IsVisible() then SetMapToCurrentZone() end
			if level==1 then
				local info;
				info = UIDropDownMenu_CreateInfo();
				info.text = L["Find Nearest"]
				info.icon = nil;
				info.arg1 = nil;
				info.isNotRadio = true;
				info.func =  nil;
				info.notCheckable = true;
				info.keepShownOnClick = false;
				info.hasArrow = true;
				info.isNotRadio = true;
				--[[info.tCoordLeft = 0;
				info.tCoordRight = 1;
				info.tCoordTop = 0;
				info.tCoordBottom = 1;]]
				UIDropDownMenu_AddButton(info, level)
			end
			if level>=2 then
				Get_FindNearestDropDown_Initialize(2, L["Find Nearest"])(info, level)
			end
			orig_MiniMapTrackingDropDown_Initialize(self, level)
		end
		DGV:RegisterEvent("WORLD_MAP_UPDATE")
		WMT:UpdateTrackingMap()
	end
	
	function WMT:Unload()
		MiniMapTrackingDropDown.initialize = orig_MiniMapTrackingDropDown_Initialize
		DGV:UnregisterEvent("WORLD_MAP_UPDATE")
	end
end
